#####
###  MailBoxer - a mailinglist/newsletter/mailarchive-framework for ZOPE
##   Copyright (C) 2002	Maik Jablonski (maik.jablonski@uni-bielefeld.de)
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
##   along with this program; if not, write to the Free Software
###  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#####

# $Id: MailBoxer.py,v 1.7 2002/10/22 21:03:35 mjablonski Exp $

import cgi, re, smtplib, rfc822, StringIO, multifile, mimetools, mimify, random, md5

from DateTime import DateTime
from Globals import InitializeClass,DTMLFile
from AccessControl import ClassSecurityInfo
from OFS.Folder import Folder
from zLOG import LOG, INFO
from MailBoxerTemplates import MailBoxerTemplates

manage_addMailBoxerForm = DTMLFile('dtml/manage_addMailBoxerForm', globals())

def manage_addMailBoxer (self, id,
				title='', 
				mailto='', 
				moderator=[],
				moderated=0,
				archived='not archived',
				mtahosts=[],
				smtphost='127.0.0.1',
				REQUEST=None
			):
	""" Add a new MailBoxer to current ObjectManager """

	mb = MailBoxer(id, title, mailto, moderator, moderated, archived, mtahosts)
	self._setObject(id,mb)
	mb=self._getOb(id)

	# Add a MailHost and a ZCatalog

	mb.manage_addFolder('archive','Archive')

	mb.manage_addProduct['MailHost'].manage_addMailHost('MailHost',smtp_host=smtphost)

	mb.manage_addProduct['ZCatalog'].manage_addZCatalog('Catalog','Catalog')
	mb.Catalog.manage_addIndex('mailDate',   'FieldIndex')
	mb.Catalog.manage_addIndex('mailFrom',   'TextIndex')
	mb.Catalog.manage_addIndex('mailSubject','TextIndex')
	mb.Catalog.manage_addIndex('mailBody',   'TextIndex')
	
	# Add dtml-templates
	
	for key in MailBoxerTemplates.keys():
		mb.addDTMLMethod(key,file=MailBoxerTemplates[key])
		
	# Redirect if requested TTW
	
	if REQUEST:
		return self.manage_main(self,REQUEST,update_menu=1)
		

class MailBoxer(Folder):
	""" MailBoxer: a folder with mail-managing-features """

	security = ClassSecurityInfo()
	
	archive_options = ['not archived', 'plain text', 'with attachments']
	
	meta_type = 'Mail Boxer'

	__ac_permissions__ = (
			      ('Add Folders', ('manage_addMail',)),
			     )

	# The _hashkey is used for the md5-checksum for verifying (un)subscriptions. 
	# It will be reset every time you restart your Zope-Server.
	
	_hashkey   = str(random.random())
	
	_properties=(
		{'id':'title',      'type':'string',   'mode':'w'},
		{'id':'mailto',     'type':'string',   'mode':'w'},
		{'id':'maillist',   'type':'lines',    'mode':'w'},
		{'id':'moderator',  'type':'lines',    'mode':'w'},
		{'id':'moderated',  'type':'boolean'  ,'mode':'w'},
		{'id':'archived',   'type':'selection','mode':'w', 'select_variable':'archive_options'},
		{'id':'subscribe',  'type':'string',   'mode':'w'},
		{'id':'unsubscribe','type':'string',   'mode':'w'},
		{'id':'mtahosts',   'type':'tokens',   'mode':'w'},
		{'id':'spamlist',   'type':'lines',    'mode':'w'},
		{'id':'atmask',     'type':'string',   'mode':'w'},
		{'id':'sniplist',   'type':'lines',    'mode':'w'},
		{'id':'catalog',    'type':'string',   'mode':'w'},
		{'id':'xmailer',    'type':'string',   'mode':'w'}
	)


	def __init__(self, id, title, mailto, moderator, moderated, archived, mtahosts):

		self.id           = id
		self.title        = title
		self.mailto       = mailto
		self.maillist     = []
		self.moderator    = moderator
		self.moderated    = moderated
		self.archived     = archived
		self.subscribe    = 'subscribe'
		self.unsubscribe  = 'unsubscribe'
		self.mtahosts	  = mtahosts
		self.spamlist     = []
		self.atmask	      = '(at)'
		self.sniplist     = [r'(\n>.*)+|(\n.>.*)+', r'(?s)\n-- .*', r'(?s)\n_+\n.*']
		self.catalog      = 'Catalog'
		self.xmailer      = 'MailBoxer'

	
	security.declarePublic('splitMail')
	def splitMail(self,Mail):
		""" return all headers and the rawBody of a mail """

		msg = mimetools.Message(StringIO.StringIO(Mail))

		mailHeader = {}
		
		for (key,value) in msg.items():
			mailHeader[key] = mimify.mime_decode_header(value)

		msg.rewindbody()
		mailBody=msg.fp.read()

		return (mailHeader,mailBody)


	security.declarePublic('securityCheck')
	def securityCheck(self,REQUEST):
		""" check for ip, x-mailer and spam """

		if self.mtahosts:
			if 'HTTP_X_FORWARDED_FOR' in self.REQUEST.environ.keys():
				REMOTE_IP = self.REQUEST.environ['HTTP_X_FORWARDED_FOR']
			else:
				REMOTE_IP = self.REQUEST.environ['REMOTE_ADDR']

			if REMOTE_IP not in self.mtahosts:			
				message = 'Host %s is not allowed' % (REMOTE_IP)
				LOG('MailBoxer', INFO,  message)
				raise(message)

		# mail already processed? avoid endless loops

		(mailHeader,mailBody)=self.splitMail(REQUEST['Mail'])

		if mailHeader.get('x-mailer')==self.xmailer:
			message = 'Mail already processed'
			LOG('MailBoxer', INFO, message)
			raise(message)

		# mail is a spam? log this and raise an error.

		for regexp in self.spamlist:
			if regexp<>'' and re.search(regexp,REQUEST['Mail']):
				message='Spam detected: %s' % regexp
				LOG('MailBoxer', INFO, message)
				raise(message)
	
	
	security.declarePrivate('_unpackMultifile')
	def _unpackMultifile(self,obj,multifile):
		""" unpack multifile and store mailBody and attachments """

		msg      = mimetools.Message(multifile)
		maintype = msg.getmaintype()
		subtype  = msg.getsubtype()
		name     = msg.getparam('name')

		# Recurse over all nested multiparts

		if maintype == 'multipart':
			multifile.push(msg.getparam('boundary'))
			multifile.readlines()
			while not multifile.last:
				multifile.next()
				self._unpackMultifile(obj,multifile)
			multifile.pop()
			return

		# Process MIME-encoded data

		plainfile = StringIO.StringIO()

		try:
			mimetools.decode(multifile,plainfile,msg.getencoding())

		# unknown or no encoding? 7bit, 8bit or whatever... copy literal

		except ValueError:
			mimetools.copyliteral(multifile,plainfile)

		# Get the plain text as the body of the mail...
   
		if maintype=='text' and subtype=='plain' and not name:
			obj.manage_changeProperties({'mailBody':plainfile.getvalue()})

		# Decode all the rest
  
		else:

			# No name? This is highly propably the html-encoded mail-body...

			if not name:
				name = '%s.%s' % (maintype,subtype)

			if self.archived=='with attachments':
	       
				id = DateTime().millis()

				# to be sure: test and search for a free id...
		
				while hasattr(obj,str("%s.%s" % (str(id),subtype))):  
					id = id + 1
			
				id = "%s.%s" % (str(id),subtype)
			
				obj.manage_addFile(id,title=name,
					file=plainfile.getvalue(),
					content_type=maintype+'/'+subtype
					)
					
		plainfile.close()


	def manage_addMail(self,Mail):
		""" store a mail & attachments in a folder and return it """

		(mailHeader,mailBody) = self.splitMail(Mail)

		# try the mail-date => take our own if it's rubbish
		
		try:
			time = DateTime(mailHeader['date'])
			if time >= DateTime():
				time = DateTime()
		except: 
			time = DateTime()
		
		# now let's create the date-path (yyyy/yyyy-mm) 

		year  = str(time.year())              # yyyy
		month = year  + '-' + str(time.mm())  # yyyy-mm

		# do we have a year folder already?

		archive = getattr(self,'archive')
		
		if not hasattr(archive,year):
		    archive.manage_addFolder(year,title=year)
		yearFolder=getattr(archive,year)

		# do we have a month folder already?

		if not hasattr(yearFolder,month):
		    yearFolder.manage_addFolder(month,title=month)
		monthFolder=getattr(yearFolder,month)

		# let's create the mailObject
		
		mailFolder = monthFolder
		
		# maybe it's a reply =>
		# then get the "replied-mail"-folder for storing the reply
		
		if mailHeader['subject'].find(':')<>-1:
			for currentFolder in monthFolder.objectValues(['Folder']):
				if currentFolder.mailSubject.strip()<>'' and mailHeader['subject'].find(currentFolder.mailSubject)<>-1:
					mailFolder=currentFolder
			  
		# search a free id for the mailobject
		
		id = time.millis()
		
		while hasattr(mailFolder,str(id)):  
		     id = id + 1

		id=str(id)

		Title= mailHeader['subject'] + ' / ' + mailHeader['from']

		mailFolder.manage_addFolder(id, title=Title)
		mailObject = getattr(mailFolder, id)

		# and now add some properties to our new mailobject 

		mailObject.manage_addProperty('mailFrom',mailHeader['from'],'string')
		mailObject.manage_addProperty('mailSubject',mailHeader['subject'],'string')
		mailObject.manage_addProperty('mailDate',time,'date')
		mailObject.manage_addProperty('mailBody',mailBody,'text')
		
		# search for mailBody and attachments
				
		self._unpackMultifile(mailObject, multifile.MultiFile(StringIO.StringIO(Mail)))

		# Index the new created mailFolder in the catalog
		
		if hasattr(self, self.catalog):
			getattr(self, self.catalog).catalog_object(mailObject)
	
		return mailObject
		
		
	security.declarePublic('manage_inboxer')
	def manage_inboxer(self, REQUEST):
		""" wrapper to post directly into archives """

		self.securityCheck(REQUEST)
		self.manage_addMail(REQUEST['Mail'])
		
				
	security.declarePublic('manage_listboxer')
	def manage_listboxer(self, REQUEST):
		""" bulks a mail to all members of the list """	

		self.securityCheck(REQUEST)

		# store mail in the archive?
		
		if self.archived <> 'not archived':
			self.manage_addMail(REQUEST['Mail'])

		(mailHeader,mailBody) = self.splitMail(REQUEST['Mail'])
		
		# get custom header & footer
	
		customHeader = self.mail_header(self, REQUEST, mail=mailHeader, body=mailBody).strip()
		customFooter = self.mail_footer(self, REQUEST, mail=mailHeader, body=mailBody).strip()
		
		# modify headers
		
		for hdrLine in customHeader.split("\n"):
			(key, value) = hdrLine.split(":",1)
			mailHeader[key.lower()] = value.strip()
		
		# create new header

		msgHeader=''
		
		for hdr in mailHeader.keys():
			msgHeader = msgHeader + "%s: %s\n" % (hdr.capitalize(),mailHeader[hdr])
		
		newMail = msgHeader+"\n"+mailBody+"\n"+customFooter

		# bulk mail to all members... return-path is first moderator

		try:
			smtpserver = smtplib.SMTP(self.MailHost.smtp_host, int(self.MailHost.smtp_port))					
			smtpserver.sendmail(self.moderator[0], self.maillist, newMail)
			smtpserver.quit()
		except:
			LOG('MailBoxer', INFO, 'SMTP-Error')


	security.declarePublic('manage_mailboxer')
	def manage_mailboxer(self,REQUEST):
		""" handles all mails and (un)subscriptions """

		self.securityCheck(REQUEST)
		
		# get email-address and subject
		
		(mailHeader,mailBody) = self.splitMail(REQUEST['Mail'])
		(RealName, RealEmail) = rfc822.parseaddr(mailHeader['from'])
		Subject               = mailHeader['subject']

		# lower-case all addresses for comparisons

		LowEmail = RealEmail.lower()
		
		MemberList=[]
		for addr in self.maillist:
			MemberList.append(addr.lower())

		ModeratorList=[]
		for addr in self.moderator:
			ModeratorList.append(addr.lower())

		# subscription? only subscribe if subscription is enabled.

		if self.subscribe<>'' and re.match('(?i)'+self.subscribe+"|.*: "+self.subscribe, Subject):
			if Subject.find(self.pin(mailHeader['from']))<>-1:
				if LowEmail not in MemberList:
					self.maillist.append(RealEmail)
					self.maillist.sort()
					self.maillist=self.maillist
			self.mail_subscribe(self, REQUEST, mail=mailHeader, body=mailBody)
			return

		# unsubscription? only unsubscribe if unsubscription is enabled...

		if self.unsubscribe<>'' and re.match('(?i)'+self.unsubscribe+"|.*: "+self.unsubscribe, Subject):
			if Subject.find(self.pin(mailHeader['from']))<>-1:
				if LowEmail in MemberList:
					index=MemberList.index(LowEmail)
					self.maillist=self.maillist[:index]+self.maillist[index+1:]
			self.mail_unsubscribe(self, REQUEST, mail=mailHeader, body=mailBody)
			return

		# message to a moderated list
		
		if self.moderated:
			# relay all mails from a moderator
			if LowEmail in ModeratorList:
				self.mail_handler(self, REQUEST, mail=mailHeader, body=mailBody)
			else:
				self.mail_moderator(self, REQUEST, mail=mailHeader, body=mailBody)
			return

		# list-traffic! relay all mails from members and moderators...
		
		if LowEmail in MemberList+ModeratorList:
			self.mail_handler(self, REQUEST, mail=mailHeader, body=mailBody)
			return

		# if all previous tests fail, it must be an unknown sender.
		
		self.mail_reply(self, REQUEST, mail=mailHeader, body=mailBody)
		return

	
	security.declarePublic('pin')
	def pin(self, mailFrom):
		""" returns the hex-digits for a randomized md5 """
		
		return md5.new(mailFrom+self._hashkey).hexdigest()
		
	
	security.declarePublic('snipMail')		
	def snipMail(self, strg):
		""" snips all regexps from strg and urlifies it """
		
		for regexp in self.sniplist:
			if regexp<>'':
				strg = re.sub(regexp,'[...]',strg)

		strg = cgi.escape(strg)
		
		# Urlifing can be done better... but it's quite good
		
		strg = re.sub(r'(?P<url>http://[-_&;,?:~=%#+/.0-9a-zA-Z]+)',r'<a href="\g<url>">\g<url></a>',strg)

		return strg.strip()


	security.declarePublic('at')
	def at(self,strg):
		""" replaces all @ with the atmask """
		
		return(strg.replace('@',self.atmask))
		

	security.declarePublic('nextObject')
	def nextObject(self,context):
		""" Returns next Folder in context or None """
	       
		parentObjects = context.aq_parent.objectIds('Folder')
		parentObjects.sort()

		if context.getId() in parentObjects:
			parentPosition = parentObjects.index(context.getId())
			
			if parentPosition+1<len(parentObjects):
				return getattr(context,str(parentObjects[parentPosition+1]))
		       
			elif context<>self and context.aq_parent<>self:
				if self.nextObject(context.aq_parent):
					objList = self.nextObject(context.aq_parent).objectValues('Folder')
					if objList:
						return objList[0]

		return None


	security.declarePublic('previousObject')
	def previousObject(self,context):
		""" Return previous Folder in context or None """
	       
		parentObjects = context.aq_parent.objectIds('Folder')
		parentObjects.sort()
 
		if context.getId() in parentObjects:
			parentPosition = parentObjects.index(context.getId())
			       
			if parentPosition > 0:
				return getattr(context,str(parentObjects[parentPosition-1]))
		       
			elif context<>self and context.aq_parent<>self:
				if self.previousObject(context.aq_parent):
					objList = self.previousObject(context.aq_parent).objectValues('Folder')
					if objList:
						return objList[-1]
		       
		return None

# And here we go to the moon... Houston, we have a problem...

InitializeClass(MailBoxer)
